import processing.core.*; 
import processing.xml.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class EvolutionEcoSystem extends PApplet {

//IMPLEMENTACI\u00d3N DE ECOSISTEMA VIRTUAL "COOPERATE vs DEFECT"
//Basado en:
//"El Dilema del Prisionero" (Teor\u00eda de Juegos) // http://prisonersdilemma.sergehelfrich.eu/
//Daniel Shiffman <http://www.shiffman.net> // Spring 2007, The Nature of Code.
//------------------------------------
//Dise\u00f1o: Miguel Angel Garc\u00eda Oliver // miguelgarciaoliver@gmail.com
//Implementaci\u00f3n T\u00e9cnica: Jimena Quejigo
//------------------------------------

World world;

public void setup() {
  size(800,800);
  background(0);
  //Inicializamos el mundo con 150 agentes
  world = new World(150);
  smooth();
}

public void draw() {
  fill(0,40);
  rect(0,0,2000,2000);
  fill(0);
  world.imprimir();
  world.run();
  
  
}



/**
IMPLEMENTACI\u00d3N DE ECOSISTEMA VIRTUAL "COOPERATE vs DEFECT"

Basado en:

- "El Dilema del Prisionero" (Teor\u00eda de Juegos) // http://prisonersdilemma.sergehelfrich.eu/
- Daniel Shiffman <http://www.shiffman.net> // Spring 2007, The Nature of Code.

------------------------------------

DESCRIPCI\u00d3N:

1. La EVOLUCI\u00d3N del ecosistema depende del estado de "SALUD" de los Agentes.
Este estado de "salud" se ve condicionado por la interacci\u00f3n entre los agentes,
y al mismo tiempo con el entorno:


  VALOR (+) "Salud" >> INTERACCI\u00d3N ENTRE AGENTES  >> JUEGO (porcentaje variable seg\u00fan estrategia)
  VALOR (-) "Salud" >> INTERACCI\u00d3N CON EL ENTORNO >> ENFERMEDAD (comer un rectangulo blanco)
  

2. La interacci\u00f3n entre agentes se da mediante el encuentro en el espacio,
momento en el que se inicializa el juego del "Dilema del Prisionero", donde cada agente adoptar\u00e1 una
estrategia predeterminada (Cooperar o Traicionar), obteniendo una puntuaci\u00f3n seg\u00fan la matriz de pagos:

                           Jugador B
                         +-----------+-----------+
               Jugador A |  Cooperar |  Desertar |
             +-----------+-----------+-----------+
             |  Cooperar |   3 , 3   |  0 , 5    |
             +-----------+-----------+-----------+
             |  Desertar |   5 , 0   |  1 , 1    |
             +-----------+-----------+-----------+

Hemos implementado 4 posibles estrategias/Comportamiento:

*   Tipo 0: COOPERAR siempre; (rojo)
*   Tipo 1: DESERTAR siempre; (verde)
*   Tipo 2: SIGUE A LOS DEM\u00c1S; (naranja) hacer lo que hizo la mayor\u00eda en la ronda anterior.
*   Tipo 3: OJO POR OJO; (azul): cooperar la primera vez, y en adelante hacer lo que
                        tu oponente hizo en la ronda anterior.

3.Existe un l\u00edmite de iteraciones del juego, equivalente al "ciclo de vida" para cada agente.
Al mismo tiempo, cada agente dispone de un algoritmo gen\u00e9tico, que mediante "mutaci\u00f3n" otorga una probabilidad
de que su descendencia herede la estrategia de sus progenitores. El fenotipo influye adem\u00e1s en
el tama\u00f1o y la velocidad de los agentes que son inversamente proporcionales.

4.El objetivo final del prototipo es visualizar en el ecosistema, aquellas
estrategias que mejor se adaptan y sobreviven en el tiempo.

------------------------------------
Dise\u00f1o: Miguel Angel Garc\u00eda Oliver // miguelgarciaoliver@gmail.com
Implementaci\u00f3n T\u00e9cnica: Jimena Quejigo
------------------------------------
*/



class Bloop {
  PVector location; // Localizacion
  PVector velocity;
 
  DNA dna;          // DNA
  float health;     // "salud"
  float xoff;       // perlin noise
  float yoff;
  
  // DNA determina: velocidad, tama\u00f1o, y estrategia de juego
  float r;
  float maxspeed;
  int comportamiento; //0 cooperan, 1 traicionan
  int tipo; //0 Siempre colaboran, 1 Siempre traicionan, 2 mayoria, 3 lo que le ha hecho el anterior.
  int njugadas;

  // Crear un agente
  Bloop(PVector l, DNA dna_) {
    location = l.get();
    velocity = new PVector(0,0);
    
    njugadas=0;
    health = 200;
    xoff = random(1000);
    yoff = random(1000);
    dna = dna_;
    maxspeed = map(dna.genes[0], 0, 1, 15, 0);
    r = map(dna.genes[0], 0, 2, 0, 40);
  
  }

  public void run() {
    update();
    borders();
    display();
  }

  // Comer comida/veneno que resta salud
  public void eat(Food f) {
    ArrayList<PVector> food = f.getFood();
    // comprobar colisi\u00f3n
    for (int i = food.size()-1; i >= 0; i--) {
      PVector foodLocation = food.get(i);
      float d = PVector.dist(location, foodLocation);
      if (d < r) {
        health -= 1.5f; 
        food.remove(i);
      }
    }
  }

  // Reproducci\u00f3n //la descendencia es una copia exacta o existe posibilidad de mutaci\u00f3n (tama\u00f1o, velocidad, estrategia)
  public Bloop reproduce(float myhealth,int numbloops) {
    
    if ((myhealth > 210) && (myhealth < 600) && (random(1) < 0.005f) && (numbloops<200)){
     
        
        DNA childDNA = dna.copy();
        
        childDNA.mutate(0.5f);
        PVector locationAux= new PVector(location.x+r,location.y+r);
        return new Bloop(locationAux, childDNA);
      
    } 
    else if((numbloops<40)&&(numbloops>30)){ 
        DNA childDNA = dna.copy();
        // Child DNA can mutate
        childDNA.mutate(0.5f);
        PVector locationAux= new PVector(location.x+r,location.y+r);
       
        return new Bloop(locationAux, childDNA);
        
    }
        
    else {
      return null;
    }
  }


  
  // Actualizar localizaci\u00f3n
 public void update() {
   float angle = random(TWO_PI);
    //movimiento perlin noise
    float vx = map(noise(xoff),0,1,-maxspeed,maxspeed);
    float vy = map(noise(yoff),0,1,-maxspeed,maxspeed);
    PVector velocity = new PVector(vx,vy);
   
    xoff += 0.01f;
    yoff += 0.01f;
    location.add(velocity);
    
    // el desplazamiento resta salud
    health -= 0.2f;
  }

  public void borders() {
    if (location.x < -r) location.x = width+r;
    if (location.y < -r) location.y = height+r;
    if (location.x > width+r) location.x = -r;
    if (location.y > height+r) location.y = -r;
  }

   /* //////////////4 ESTRATEGIAS POSIBLES//////////////////////////
    
   Tipo 0: COOPERAR siempre; (rojo)
   Tipo 1: DESERTAR siempre; (verde)
   Tipo 2: SIGUE A LOS DEM\u00c1S; (naranja)
   Tipo 3: OJO POR OJO; (azul)
    */
    
  public void display() {
    World w;
    
    rectMode(CENTER);    
    if(dna.genes[1] < 0.25f)
      {
        
        stroke(255,random(100),random(100),health);
        strokeWeight (1);
        //colaboran rojo
        fill(0,0);
        comportamiento=0;
        tipo=0;
      }
      
      if((dna.genes[1] >= 0.25f) &&(dna.genes[1] < 0.5f))
      {
        stroke(random(100),255,random(100),health);
        strokeWeight (1);
        fill(0,0);
        //traicionan verde
        comportamiento=1;
        tipo=1;
      }
      
      if((dna.genes[1] >= 0.5f) &&(dna.genes[1] < 0.75f))
      {
         stroke(245,165,5,health); //naranja
        strokeWeight (1);
        fill(0,0);
        //lo que hace la mayoria, se inicializa con random
        comportamiento=PApplet.parseInt(random(0,2));//es entre 0 y 2 porque trunca los valores
        tipo=2;
      }
      
      if((dna.genes[1] >= 0.75f) &&(dna.genes[1] <= 1))
      {
        stroke(39,207,203,health); //Azul
        strokeWeight (1);
        fill(0,0);
       //lo que le ha hecho el ultimo. inicialmente colaboran
        comportamiento=0;
        tipo=3;
      }
    
    
    rect(location.x, location.y, r, r); 
    
  }

  //////////// Mueren cuando la salud es 0
  
  public boolean dead() {
    if ((health < 0.0f)) {
      return true;
    } 
    else {
      return false;
    }
  }
}



class DNA {

  float[] genes;
 
  
  // Constructor
  DNA() {
    //random 0 y 1 (!!)
    genes = new float[2];
    for (int i = 0; i < genes.length; i++) {
      genes[i] = random(0,1);
    }   
  }
  
  DNA(float[] newgenes) {
    genes = newgenes;
  }
  
  public DNA copy() {
    float[] newgenes = new float[genes.length];
    arraycopy(genes,newgenes);
    return new DNA(newgenes);
  }
  
  // Mutacion tama\u00f1o, velocidad, y estrategia
  public void mutate(float m) {
    for (int i = 0; i < genes.length; i++) {
      if (random(0,1) <= m) {
         genes[i] = random(0,1);
      }
    }
  }
}


class Food {
  ArrayList<PVector> food;
 
  Food(int num) {
    
    food = new ArrayList();
    for (int i = 0; i < num; i++) {
       food.add(new PVector(random(width),random(height))); 
    }
    
  } 
  
  // A\u00f1adir comida
  public void add(PVector l) {
     food.add(l.get()); 
  }
  
  // Display comida
  public void run() {
    for (PVector f : food) {
       rectMode(CENTER);
       stroke(0);
       fill(175);
       rect(f.x,f.y,8,8);
    } 
    
    // la comida/veneno aparece de manera aleatoria
    if (random(1) < 0.0005f) {
       food.add(new PVector(random(width),random(height))); 
    }
  }
  
  public ArrayList getFood() {
    return food; 
  }
}


class World {

  ArrayList<Bloop> bloops;
  Food food;

  // Constructor
  World(int num) {
    
    food = new Food(PApplet.parseInt(random(num)));
    bloops = new ArrayList<Bloop>();
    for (int i = 0; i < num; i++) {
      PVector l = new PVector(random(width),random(height));
      DNA dna = new DNA();
      bloops.add(new Bloop(l,dna));
    }
  }

  // Nacimiento de agentes
  public void born(float x, float y) {
    PVector l = new PVector(x,y);
    DNA dna = new DNA();
    bloops.add(new Bloop(l,dna));
  }

  // Correr el mundo
  public void run() {
    food.run();
   
    
   
    for (int i = bloops.size()-1; i >= 0; i--) {
      // Los Agentes se mueven e interaccionan con el entorno
      Bloop b = bloops.get(i);
      b.run();
      b.eat(food);
      
      //Comprobamos si colisiona
     for (int j = bloops.size()-1; j >= 0; j--) {
        Bloop b1 = bloops.get(j);
        
        if(i!=j){
          
          float d = PVector.dist(b.location,b1.location);//Distancia entre 2 Agentes
          
          if ((d < b.r/3) || (d < b1.r/3)) { println("JUEGO1");Jugar(i,j);println("JUEGO2");} ////Si han colisionado jugaran
           }
        
      }
      
      // Si muere un agente, se elimina y se crea m\u00e1s comida/veneno
     if (b.dead()) {
        bloops.remove(i);
        food.add(b.location);
      }
      
      // Probabilidad de reproducirse
      Bloop child = b.reproduce(b.health,bloops.size());
      if (child != null) bloops.add(child);
   
  }
}

//////////JUEGO DEL DILEMA DEL PRISIONERO /////////////

  public void Jugar(int i, int j)
  {
    Bloop b = bloops.get(i);
    Bloop b1 = bloops.get(j);
    
   //JUEGAN SI SU NUMERO DE PARTIDAS ES MENOR QUE 40
   //edad en la que se jubilan.//Esto da equilibrio al ecosistema.
   // CASOS POSIBLES: 
   
    if((b.njugadas<=40)||(b1.njugadas<=40)){
     //Los 2 colaboran
     if((b.comportamiento==0) && (b1.comportamiento==0))
        { 
          b.health += 3; b1.health +=3;//Actualizamos la salud
          ModificarComportamiento(i,j);           
          }
           
      //b colabora y b1 traiciona
      if((b.comportamiento==0) && (b1.comportamiento==1)) 
         {
           b.health += 0; b1.health +=4;//Actalizamos salud
           ModificarComportamiento(i,j);              
          }
            
        //Los b traiciona y b1 colabora
        if((b.comportamiento==1) && (b1.comportamiento==0)) 
           {
             b.health += 4; b1.health +=0;//Actualizamos salud
             ModificarComportamiento(i,j);
           }
            
        //Los 2 traicionan
        if((b.comportamiento==1) && (b1.comportamiento==1)) 
            {
              b.health += 1; b1.health +=1;//Actualizamos salud
              ModificarComportamiento(i,j);
            } 
             //incrementamos las jugadas
    b.njugadas++;
    b1.njugadas++;
  }//if((b.njugadas<=30)||(b1.njugadas<=30))    
         
  }

//Modifica el comportamiento de los bloops
  public void ModificarComportamiento(int i, int j){
      //variable auxiliar para guardar el comportamiento, porque en el caso de que ambos agentes sean tipo 3 hay que guardar
       //en comportamiento del primero que se cambie para que luego lo tome el otro
       
     int AuxComportamiento; 
     boolean b1Cambio=false;
     Bloop b = bloops.get(i);
     Bloop b1 = bloops.get(j);
     
       if(b.tipo==2){ //comportamiento que adopta la mayor\u00eda 
             AuxComportamiento=b.comportamiento;
             b.comportamiento= mayoria();
             
             if(b1.tipo==2){b1.comportamiento=mayoria(); b1Cambio=true;}
             if(b1.tipo==3){ b1.comportamiento=AuxComportamiento; b1Cambio=true;}     
        }//if(b.tipo==2)
        
        if(b.tipo==3){
          AuxComportamiento=b1.comportamiento;
          
          if(b1.tipo==2){b1.comportamiento=mayoria();  b1Cambio=true;}
          if(b1.tipo==3){ b1.comportamiento=b.comportamiento; b1Cambio=true;}
        
          b.comportamiento=AuxComportamiento;  
        }
        
        if(b1Cambio==false){//Si b1 no ha cambiado
          if(b1.tipo==2){
            AuxComportamiento=b1.comportamiento;
            b1.comportamiento= mayoria();
            
            if(b.tipo==2){b.comportamiento=mayoria();}
            if(b.tipo==3){b.comportamiento=AuxComportamiento;}  
          }
          
          if(b1.tipo==3){ //comportamiento que adopta el anterior
               AuxComportamiento=b.comportamiento;  
               
               if(b.tipo==2){b.comportamiento=mayoria();}
               if(b.tipo==3){b.comportamiento=b1.comportamiento;}  
              
               b1.comportamiento=AuxComportamiento;   
          }//if(b1.tipo==3)
        }
    
}//void ModificarComportamiento(int i, int j)

 //Obtiene la mayoria de comportamientos de los agentes
 public int mayoria()
    {
    int contCol=0; //Contador colaboraci\u00f3n
    int contTra=0; // Contador traici\u00f3n
    
     for (int i = bloops.size()-1; i >= 0; i--) {
        Bloop b = bloops.get(i);
        if(b.comportamiento==0){contCol++;}
        else {contTra++;}
     }
     //Devolvemos la mayoria
     if (contCol>=contTra){return 0;}
     else {return 1;}
    }
    
 public void imprimir(){
  int colaboran=0;
  int traicionan=0;
  int mayoria=0;
  int anterior=0;
  float salud=0.0f;
  
  int colaborativos=0;
  int traidores=0;
  
  for (int i = bloops.size()-1; i >= 0; i--) {
    Bloop b = bloops.get(i);
    salud+=b.health;
    if(b.tipo==0){colaboran++;}
    if(b.tipo==1){traicionan++;}
    if(b.tipo==2){mayoria++;}
    if(b.tipo==3){anterior++;}
    
    if(b.comportamiento==0){colaborativos++;}
    if(b.comportamiento==1){traidores++;}
  }
  
  salud=salud/bloops.size();
   
  fill(255);
  
  text("Average health: " + salud, 20, 18);
  text("------------------",20, 36);
  noStroke();
  fill(255,0,0);
  rect(25,49,10,10);
  fill(255);
  text("Always cooperate: " + (colaboran*100)/bloops.size() +" %", 35, 54);
  noStroke();
  fill(0,255,0);
  rect(25,67,10,10);
  fill(255);
  text("Always defect: " + (traicionan*100)/bloops.size() +" %", 35, 72);
  noStroke();
  fill(245,165,5);
  rect(25,85,10,10);
  fill(255);
  text("Follow others: " + (mayoria*100)/bloops.size() +" %", 35, 90);
  noStroke();
  fill(39,207,203);
  rect(25,103,10,10);
  fill(255);
  text("Tit for Tat: " + (anterior*100)/bloops.size() +" %", 35,108);
  
   
 }
    
}//Termina la clase




  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "EvolutionEcoSystem" });
  }
}
